@@ -3,10 +3,8 @@
 import numpy as np
 import h5py
 import yaml
-import yamlloader
 import os
 
-from collections import OrderedDict
 from collections import abc
 
 from sympy import Integer
@@ -20,7 +18,7 @@ from sympy.core.expr import AtomicExpr
 
 from sympde.core.basic import CalculusFunction
 from .basic            import BasicDomain, InteriorDomain, Boundary, Union, Connectivity
-from .basic            import Interval, Interface
+from .basic            import Interval, Interface, CornerBoundary, CornerInterface
 from .basic            import ProductDomain
 
 # TODO fix circular dependency between domain and mapping
@@ -75,7 +73,6 @@ class Domain(BasicDomain):
             interiors = Tuple(*interiors)
         # ...
 
-        # ...
         if not( boundaries is None ):
             if not isinstance( boundaries, (*iterable_types, Boundary)):
                 raise TypeError('> Expecting an iterable or a Boundary')
@@ -91,9 +88,7 @@ class Domain(BasicDomain):
             boundaries = []
 
         boundaries = Tuple(*boundaries)
-        # ...
 
-        # ...
         if not( connectivity is None ):
             if not isinstance( connectivity, Connectivity ):
                 raise TypeError('> Expecting a Connectivity')
@@ -122,6 +117,7 @@ class Domain(BasicDomain):
 
         obj = Basic.__new__(cls, name, interiors, boundaries, mapping)
         obj._connectivity   = connectivity
+        obj._corners        = None
         obj._dtype          = dtype
         obj._dim            = dim
         obj._logical_domain = logical_domain
@@ -163,6 +159,14 @@ class Domain(BasicDomain):
     def interfaces(self):
         return self.connectivity.interfaces
 
+    @property
+    def corners(self):
+        corners = getattr(self,'_corners', None)
+        if corners is None:
+            corners = self.get_shared_corners()
+        self._corners = corners
+        return corners
+
     def __len__(self):
         if isinstance(self.interior, InteriorDomain):
             return 1
@@ -179,58 +183,46 @@ class Domain(BasicDomain):
             return [i.name for i in self.interior.args]
 
 
+    def set_interfaces(self, *interfaces):
+        for i in interfaces:
+            self.connectivity[i.name] = i
+
     def _sympystr(self, printer):
         sstr = printer.doprint
-        if self.mapping:
-            return '{}({})'.format(sstr(self.mapping), sstr(self.name))
-        else:
-            return '{}'.format(sstr(self.name))
+        return '{}'.format(sstr(self.name))
 
-    def get_boundary(self, name=None, axis=None, ext=None):
+    def get_boundary(self, axis, ext):
         """return boundary by name or (axis, ext)."""
         # ...
-        if name is None:
-            assert(not( ext  is None ))
 
-            if axis is None:
-                assert(self.interior.dim == 1)
-                axis = 0
-
-            if axis == 0:
-                if ext == -1:
-                    name = r'\Gamma_1'
-
-                if ext == 1:
-                    name = r'\Gamma_2'
-
-            if axis == 1:
-                if ext == -1:
-                    name = r'\Gamma_3'
-
-                if ext == 1:
-                    name = r'\Gamma_4'
-
-            if axis == 2:
-                if ext == -1:
-                    name = r'\Gamma_5'
-
-                if ext == 1:
-                    name = r'\Gamma_6'
+        if axis is None:
+            assert(self.interior.dim == 1)
+            axis = 0
         # ...
 
         if isinstance(self.boundary, Union):
-            x = [i for i in self.boundary.args if i.name == name]
+            x = [i for i in self.boundary.args if i.ext == ext and i.axis == axis]
             if len(x) == 0:
-                raise ValueError('> could not find boundary {}'.format(name))
+                raise ValueError('> could not find boundary with axis {} and ext {}'.format(axis, ext))
 
             return x[0]
 
         elif isinstance(self.boundary, Boundary):
-            if self.boundary.name == name:
+            if self.boundary.axis == axis and self.boundary.ext == ext:
                 return self.boundary
 
-            else:
-                return None
+        raise ValueError('> could not find boundary with axis {} and ext {}'.format(axis, ext))
+
+    def get_interface(self, domain1, domain2):
+        interfaces = []
+        for i in self.interface:
+            if i.plus in [domain1, domain2]:
+                if i.minus in [domain1, domain2]:
+                    interfaces.append(i)
+        if interfaces:
+            return Union(*interfaces)
+
+        raise ValueError('> could not find the interface of {} and {}'.format(domain1, domain2))
 
     def get_interior(self, name):
         """return interior by name."""
@@ -266,7 +258,7 @@ class Domain(BasicDomain):
              'boundary':     boundary,
              'connectivity': connectivity}
 
-        return OrderedDict(sorted(d.items()))
+        return dict(sorted(d.items()))
 
     def export( self, filename ):
 
@@ -274,7 +266,7 @@ class Domain(BasicDomain):
 
         # Dump metadata to string in YAML file format
         geo = yaml.dump( data   = yml,
-                         Dumper = yamlloader.ordereddict.Dumper )
+                         sort_keys = None)
 
         # Create HDF5 file (in parallel mode if MPI communicator size > 1)
         h5 = h5py.File( filename, mode='w' )
@@ -289,7 +281,8 @@ class Domain(BasicDomain):
     def from_file( cls, filename ):
 
         # ... check extension of the file
-        basename, ext = os.path.splitext(filename)
+        _, ext = os.path.splitext(filename)
+
         if not(ext == '.h5'):
             raise ValueError('> Only h5 files are supported')
         # ...
@@ -355,33 +348,27 @@ class Domain(BasicDomain):
             connectivity[edge] = Interface(edge, bnds[0], bnds[1])
         # ...
 
+        logical_domain = Domain(domain_name, dim=dim,
+                                interiors=interior,
+                                boundaries=boundary,
+                                connectivity=connectivity)
+
         obj = Domain.__new__(cls, domain_name,
                               interiors=interior,
                               boundaries=boundary,
-                              connectivity=connectivity)
-        return mapping(obj)
+                              connectivity=connectivity,
+                              mapping = mapping, logical_domain=logical_domain)
+        return obj
 
-    def join(self, other, name, bnd_minus=None, bnd_plus=None):
+    def join(self, other, name, bnd_minus=None, bnd_plus=None, direction=None):
 
-        from sympde.topology.mapping import InterfaceMapping, MultiPatchMapping
+        from sympde.topology.mapping import MultiPatchMapping
         # ... connectivity
         connectivity = Connectivity()
-        # TODO be careful with '|' in psydac
+
         if bnd_minus and bnd_plus:
-
-            if bnd_minus.mapping and bnd_plus.mapping:
-                int_map            = InterfaceMapping(bnd_minus.mapping , bnd_plus.mapping)
-                a,b                = bnd_minus.logical_domain, bnd_plus.logical_domain
-                l_name             = '{l}|{r}'.format(l=a.domain.name, r=b.domain.name)
-                int_logical_domain = Interface(l_name, a,b)
-            else:
-                int_map            = None
-                int_logical_domain = None
-
-            name               = '{l}|{r}'.format(l=bnd_minus.domain.name, r=bnd_plus.domain.name)
-            connectivity[name] = Interface(name, bnd_minus, bnd_plus,
-                                           mapping=int_map,
-                                           logical_domain=int_logical_domain)
+            interface          = bnd_minus.join(bnd_plus, direction=direction)
+            connectivity[interface.name] = interface
 
         for k,v in self.connectivity.items():
             connectivity[k] = v
@@ -390,10 +377,7 @@ class Domain(BasicDomain):
             connectivity[k] = v
 
         # ... boundary
-        boundaries_minus = self.boundary.complement(bnd_minus)
-        boundaries_plus  = other.boundary.complement(bnd_plus)
-
-        boundaries = Union(boundaries_minus, boundaries_plus)
+        boundaries = Union(self.boundary, other.boundary).complement(Union(bnd_minus, bnd_plus))
         boundaries = boundaries.as_tuple()
 
         # ... interiors
@@ -422,6 +406,61 @@ class Domain(BasicDomain):
                       mapping=mapping,
                       logical_domain=logical_domain)
 
+    def get_shared_corners(self):
+        """ Compute the corners shared by multiple patches in 2D """
+
+        interfaces   = self.interfaces
+        interfaces = (interfaces,) if isinstance(interfaces, Interface) else interfaces
+
+        directions   = {i.plus:i.direction for i in interfaces}
+        directions.update({i.minus:i.direction for i in interfaces})
+
+        boundaries    = {i.minus:i.plus for i in interfaces}
+        boundaries.update({value:key for key, value in boundaries.items()})
+
+        not_treated_corners = set([tuple(set((b, n))) for b in boundaries for n in b.adjacent_boundaries])
+        grouped_corners     = []
+
+        while not_treated_corners:
+            corner = not_treated_corners.pop()
+            grouped_corners.append([corner])
+            if not ( corner[0] in boundaries and corner[1] in boundaries):
+                while corner[1] in boundaries:
+                    bd1     = boundaries[corner[1]]
+                    bd2     = bd1.domain.get_boundary(axis=corner[0].axis, ext=corner[0].ext)
+                    corner = (bd1, bd2.rotate(directions[bd1]))
+                    grouped_corners[-1].append(corner)
+
+                corner = grouped_corners[-1][0]
+                while corner[0] in boundaries:
+                    bd2     = boundaries[corner[0]]
+                    bd1     = bd2.domain.get_boundary(axis=corner[1].axis, ext=corner[1].ext)
+                    corner = (bd1.rotate(directions[bd2]), bd2)
+                    grouped_corners[-1].insert(0, corner)
+
+            else:
+                while corner[1] in boundaries:
+                    bd1     = boundaries[corner[1]]
+                    bd2     = bd1.domain.get_boundary(axis=corner[0].axis, ext=corner[0].ext)
+                    corner = (bd1, bd2.rotate(directions[bd1]))
+                    if corner == grouped_corners[-1][0]:
+                        break
+                    grouped_corners[-1].append(corner)
+                else:
+                    corner = grouped_corners[-1][0]
+                    while corner[0] in boundaries:
+                        bd2     = boundaries[corner[0]]
+                        bd1     = bd2.domain.get_boundary(axis=corner[1].axis, ext=corner[1].ext)
+                        corner = (bd1.rotate(directions[bd2]), bd2)
+                        grouped_corners[-1].insert(0, corner)
+
+            grouped_corners[-1] = tuple(tuple(set(c)) for c in grouped_corners[-1])
+            not_treated_corners = not_treated_corners.difference(grouped_corners[-1])
+
+        grouped_corners = set(tuple(grouped_corners))
+        grouped_corners = Union(*[CornerInterface(*[CornerBoundary(*e) for e in cs]) for cs in grouped_corners])
+        return grouped_corners
+
 #==============================================================================
 class PeriodicDomain(BasicDomain):
 
@@ -476,8 +515,22 @@ class NCubeInterior(InteriorDomain):
 
         obj = InteriorDomain.__new__(cls, name, dim=dim, dtype=dtype,
                     mapping=mapping, logical_domain=logical_domain)
+
         obj._min_coords = min_coords
         obj._max_coords = max_coords
+
+        boundaries = []
+        i = 1
+        for axis in range(dim):
+            for ext in [-1, 1]:
+                bnd_name = r'\Gamma_{}'.format(i)
+                bd_logical_domain = logical_domain
+                if bd_logical_domain:
+                    bd_logical_domain = bd_logical_domain.get_boundary(axis=axis, ext=ext)
+                Gamma = Boundary(bnd_name, obj, axis=axis, ext=ext, mapping=mapping, logical_domain=bd_logical_domain)
+                boundaries += [Gamma]
+                i += 1
+        obj._boundary   = Union(*boundaries)
         return obj
 
     @property
@@ -488,6 +541,23 @@ class NCubeInterior(InteriorDomain):
     def max_coords(self):
         return self._max_coords
 
+    @property
+    def boundary(self):
+        return self._boundary
+
+    def get_boundary(self, axis=None, ext=None):
+        """return boundary by (axis, ext)."""
+        # ...
+        assert(not( ext  is None ))
+
+        if axis is None:
+            assert(self.dim == 1)
+            axis = 0
+
+        if isinstance(self.boundary, Union):
+            x = [i for i in self.boundary.args if i.ext == ext and i.axis==axis]
+            if x:return x[0]
+        raise ValueError('> could not find boundary {}'.format(name))
 #==============================================================================
 # Ncube's properties (in addition to Domain's properties):
 #   . min_coords (default value is tuple of zeros)
@@ -532,6 +602,9 @@ class NCube(Domain):
         #   cls = globals()[dtype['type']]
         #   domain = cls(name, **dtype['parameters'])
         #
+        min_coords = tuple(float(i) for i in min_coords)
+        max_coords = tuple(float(i) for i in max_coords)
+
         if dim == 1:
             cls = Line
             dtype = {'type': 'Line',
@@ -558,20 +631,10 @@ class NCube(Domain):
         else:
             interior = ProductDomain(*intervals, name=name)
 
-        interior = NCubeInterior(interior, dtype=dtype, min_coords=tuple(min_coords), max_coords=tuple(max_coords))
-
-        boundaries = []
-        i = 1
-        for axis in range(dim):
-            for ext in [-1, 1]:
-                bnd_name = r'\Gamma_{}'.format(i)
-                Gamma = Boundary(bnd_name, interior, axis=axis, ext=ext)
-                boundaries += [Gamma]
-                i += 1
+        interior = NCubeInterior(interior, dim=dim, dtype=dtype, min_coords=tuple(min_coords), max_coords=tuple(max_coords))
 
         # Create instance of given type
-        obj = super().__new__(cls, name, interiors=[interior],
-                boundaries=boundaries)
+        obj = super().__new__(cls, name, interiors=[interior], boundaries=interior.boundary)
 
         # Store attributes in object
         obj._coordinates = tuple(coordinates)
